/*
 * Copyright (c) 2023 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

#include <errno.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <zephyr/kernel.h>
#include <zephyr/types.h>
#include <zephyr/logging/log.h>
LOG_MODULE_REGISTER(home_demo, LOG_LEVEL_INF);
#include <zephyr/net/wifi.h>
#include <zephyr/net/wifi_mgmt.h>
#include <zephyr/net/net_mgmt.h>
#include <net/wifi_mgmt_ext.h>
#include <net/wifi_credentials.h>
#include <zephyr/net/socket.h>

/* STEP 1.2 - Include the header file of the HTTP client library */
#include <zephyr/net/http/client.h>

// header files for uart 
#include "uart_async_adapter.h"
#include <zephyr/drivers/uart.h>
#include <zephyr/usb/usb_device.h>

#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <soc.h>
#include <zephyr/settings/settings.h>

// #include <cJSON.h>
// #include <cJSON_os.h>

#include <cType.h>
#include <zephyr/sys/printk.h>
#include <zephyr/sys/byteorder.h>
#include <math.h>
#include <time.h>

//for sd card
#include <zephyr/storage/disk_access.h>
#include <zephyr/fs/fs.h>
#include <ff.h>
// #include <zephyr/sys/time_units.h>
// #include <zephyr/sys_clock.h>

#include <zephyr/drivers/gpio.h>

//for microphone
#include <zephyr/zephyr.h>
#include <nrfx_pdm.h>
#include <hal/nrf_gpio.h>
#include <zephyr/sys/mutex.h>

#include <zephyr/net/tls_credentials.h>





#define CONTAINER_OF(ptr, type, field) ((type *)(((char *)(ptr)) - offsetof(type, field)))

#define EVENT_MASK (NET_EVENT_L4_CONNECTED | NET_EVENT_L4_DISCONNECTED)


/* STEP 2 - Define the macros for the HTTP server hostname and port */
#define HTTPS_HOSTNAME "storage.googleapis.com"
#define HTTPS_PORT "443"

/* STEP 4.1 - Define a macro for the credential security tag */
#define HTTP_TLS_SEC_TAG 1


/* STEP 2.3 - Include the certificate */
// static const char ca_certificate[] = {
// 	#include "certificate.h"
// };
static const char cert[] = {
	#include "certificate.pem.inc"

	/* Null terminate certificate if running Mbed TLS on the application core.
	 * Required by TLS credentials API.
	 */
	IF_ENABLED(CONFIG_TLS_CREDENTIALS, (0x00))
};

// static const unsigned char private_key[] = {
//     #include "private_key.pem.inc"
// };



/* STEP 3 - Declare the necessary buffers for receiving messages */
#define RECV_BUF_SIZE 2048
#define CLIENT_ID_SIZE 36

// uart macros
#define UART_BUF_SIZE 813
#define UART_WAIT_FOR_BUF_DELAY K_MSEC(50)
#define UART_WAIT_FOR_RX 50000


#define LED1_NODE	DT_ALIAS(led1)
#define LED2_NODE	DT_ALIAS(led2)
#define LED3_NODE	DT_ALIAS(led3)
#define SW1_NODE	DT_ALIAS(button1)
#define SW2_NODE	DT_ALIAS(button2)

#define PDM_BUF_SIZE 1024
#define CLK_PIN 24 // change for specific pin
#define DIN_PIN 27 // change for specific pin

// #define QUEUE_SIZE 100 // Size of the queue

// #define DOUBLE_BUFFER_SIZE PDM_BUF_SIZE

// static int16_t pcm_double_buffer[2][DOUBLE_BUFFER_SIZE];
// static volatile bool active_buffer = 0;
// static volatile bool buffer_ready[2] = {false, false};
#define CIRCULAR_BUFFER_SIZE 64 // Size of the circular buffer

// static int16_t circular_buffer[CIRCULAR_BUFFER_SIZE][PDM_BUF_SIZE];
static int16_t *circular_buffer = NULL;
static volatile int head = 0;
static volatile int tail = 0;
static volatile bool buffer_full = false;
static volatile bool buffer_ready = false;

static char recv_buf[RECV_BUF_SIZE];
static char client_id_buf[CLIENT_ID_SIZE+2];

/* STEP 4 - Define the variable for the counter as 0 */
static int counter = 0;

static int sock;
static struct sockaddr_storage server;

static struct net_mgmt_event_callback mgmt_cb;
static bool connected_wifi;
static K_SEM_DEFINE(run_app, 0, 1);

// semaphore of sd card

// K_SEM_DEFINE(sd_read_sem,0,1);
// K_SEM_DEFINE(sd_write_sem,0,1);

K_MUTEX_DEFINE(file_mutex);
// K_SEM_DEFINE(storing_func_sem_2,0,1);

// structures of uart
static const struct device *uart = DEVICE_DT_GET(DT_NODELABEL(uart0));				//*uart is responsible to handle uart api
static struct k_work_delayable uart_work;


// struct k_work_delayable check_cond_work;

struct uart_data_t {
	void *fifo_reserved;
	uint8_t data[UART_BUF_SIZE];
	uint16_t len;
};

typedef struct {
    int prodotem;
	int prodohumi;
	int prodonoi;
	int prodop1;
	int prodop10;
	int prodop25;
	int prodoco2;
	float prodohch;
	int prodovo;
	int prodoco;
	int prodored;
	int prodogreen;
	int prodoblue;
	float prodoUVa;
	float prodoUVb;
	float prodoUVc;
	int prodoir;
} SensorData;

struct wav_header {
    char riff_header[4];       // "RIFF"
    uint32_t wav_size;         // File size - 8
    char wave_header[4];       // "WAVE"
    char fmt_header[4];        // "fmt "
    uint32_t fmt_chunk_size;   // 16 for PCM
    uint16_t audio_format;     // 1 for PCM
    uint16_t num_channels;     // 1 for mono, 2 for stereo
    uint32_t sample_rate;      // Sampling frequency
    uint32_t byte_rate;        // sample_rate * num_channels * (bits_per_sample / 8)
    uint16_t block_align;      // num_channels * (bits_per_sample / 8)
    uint16_t bits_per_sample;  // Bits per sample
    char data_header[4];       // "data"
    uint32_t data_bytes;       // Number of bytes in data
} __attribute__((packed));

// variables of uart
static K_FIFO_DEFINE(fifo_uart_tx_data);						//used to define and initialize a FIFO
static K_FIFO_DEFINE(fifo_uart_rx_data);

// functions of uart
bool dataReceived;
// char *data_over_wifi = NULL;
struct uart_data_t *buf;
char data_wifi[275];


char* process_uart_data(const char *bufferData);
char* removeWhitespace(const char* json_data);

static time_t m_time, m_last_calibrate_time;
struct tm       *local_time_now;

time_t newtime;

//SD CARD
static int lsdir(const char *path);
static int write_file(const char *filename, const char *data);
static int read_file(const char *filename);
static int clear_file(const char *filename);

//gpio spec get
static const struct gpio_dt_spec led1 = GPIO_DT_SPEC_GET(LED1_NODE, gpios);
static const struct gpio_dt_spec led2 = GPIO_DT_SPEC_GET(LED2_NODE, gpios);
static const struct gpio_dt_spec led3 = GPIO_DT_SPEC_GET(LED3_NODE, gpios);
static const struct gpio_dt_spec sw1 = GPIO_DT_SPEC_GET(SW1_NODE, gpios);
static const struct gpio_dt_spec sw2 = GPIO_DT_SPEC_GET(SW2_NODE, gpios);

static struct gpio_callback sw1_cb_data;
static struct gpio_callback sw2_cb_data;

static uint8_t cloud_flag;

static FATFS fat_fs;
/* mounting info */
static struct fs_mount_t mp = {
    .type = FS_FATFS,
    .fs_data = &fat_fs,
};


/*
 *  Note the fatfs library is able to mount only strings inside _VOLUME_STRS
 *  in ffconf.h
 */
static const char *disk_mount_pt = "/SD:";
// const char *filename = "/SD:/home.json";
const char *wav_filename = "/SD:/five.wav";


enum cloudStatus{
	no_data_from_prodo = 2,
	send_data_from_prodo = 1,
	wifi_disconnected = 5
};

int16_t pdm_buf[PDM_BUF_SIZE];

// Flags for timer
uint8_t five_min_over_flag = 0;
uint8_t http_req_success_flag = 0;

#define FILENAME_TEMPLATE "/SD:/%i.json"
#define MAX_FILENAME_LENGTH 255

int current_file_number = 1;
char filename[MAX_FILENAME_LENGTH];
char full_filename[MAX_FILENAME_LENGTH];

static int cmd_wifi_status(void);

/* Provision certificate to modem */
// int cert_provision(void)
// {
// 	int err;

// 	printk("Provisioning certificate\n");

// #if CONFIG_MODEM_KEY_MGMT
// 	bool exists;
// 	int mismatch;

// 	/* It may be sufficient for you application to check whether the correct
// 	 * certificate is provisioned with a given tag directly using modem_key_mgmt_cmp().
// 	 * Here, for the sake of the completeness, we check that a certificate exists
// 	 * before comparing it with what we expect it to be.
// 	 */
// 	err = modem_key_mgmt_exists(TLS_SEC_TAG, MODEM_KEY_MGMT_CRED_TYPE_CA_CHAIN, &exists);
// 	if (err) {
// 		printk("Failed to check for certificates err %d\n", err);
// 		return err;
// 	}

// 	if (exists) {
// 		mismatch = modem_key_mgmt_cmp(TLS_SEC_TAG, MODEM_KEY_MGMT_CRED_TYPE_CA_CHAIN, cert,
// 					      sizeof(cert));
// 		if (!mismatch) {
// 			printk("Certificate match\n");
// 			return 0;
// 		}

// 		printk("Certificate mismatch\n");
// 		err = modem_key_mgmt_delete(TLS_SEC_TAG, MODEM_KEY_MGMT_CRED_TYPE_CA_CHAIN);
// 		if (err) {
// 			printk("Failed to delete existing certificate, err %d\n", err);
// 		}
// 	}

// 	printk("Provisioning certificate\n");

// 	/*  Provision certificate to the modem */
// 	err = modem_key_mgmt_write(TLS_SEC_TAG, MODEM_KEY_MGMT_CRED_TYPE_CA_CHAIN, cert,
// 				   sizeof(cert) - 1);
// 	if (err) {
// 		printk("Failed to provision certificate, err %d\n", err);
// 		return err;
// 	}
// #else /* CONFIG_MODEM_KEY_MGMT */
// 	err = tls_credential_add(TLS_SEC_TAG,
// 				 TLS_CREDENTIAL_CA_CERTIFICATE,
// 				 cert,
// 				 sizeof(cert));
// 	if (err == -EEXIST) {
// 		printk("CA certificate already exists, sec tag: %d\n", TLS_SEC_TAG);
// 	} else if (err < 0) {
// 		printk("Failed to register CA certificate: %d\n", err);
// 		return err;
// 	}
// #endif /* !CONFIG_MODEM_KEY_MGMT */

// 	return 0;
// }


/* Timer callback function */
void timer_callback(struct k_timer *timer_id)
{
    LOG_INF("5-minute timer expired!\n");
	five_min_over_flag = 1;
	// k_mutex_lock(&file_mutex, K_FOREVER);
	// read_file(filename);
	// // client_get_new_id();
	// clear_file(filename);
	// k_mutex_unlock(&file_mutex);
	/* Restart the timer */
    // k_timer_start(timer_id, K_MSEC(300000), K_NO_WAIT);
}
/* Define and initialize the timer */
K_TIMER_DEFINE(five_minute_timer, timer_callback, NULL);

void nrf_cal_set_time(time_t epoch)
{
    time_t uncal_difftime, difftime, newtime;
    newtime = epoch;
    // LOG_INF("newtime epoch %ld",newtime);    
    m_last_calibrate_time = k_uptime_get();         //Kernel time at the time of epoch write
    m_time = newtime;
    LOG_INF(" cal set time: %d,%d,%d\n",m_time,m_last_calibrate_time,newtime);
}    

time_t nrf_cal_get_time(void)
{
    int64_t return_time;
    // return_time = m_time + ((k_uptime_get()-m_last_calibrate_time))/1000;
	// LOG_INF("Return time: %d",return_time);
	return_time = k_uptime_get()/1000;
	return return_time;
	
}

time_t nrf_cal_get_time_calibrated(void)
{
    return nrf_cal_get_time();
}

static void uart_cb(const struct device *dev, struct uart_event *evt, void *user_data)
{
	ARG_UNUSED(dev);									//Mark the unused parameter to avoid compiler warnings

	static size_t aborted_len;
	
	static uint8_t *aborted_buf;
	static bool disable_req = false;

	// below mentioned cases are the uart events
	switch (evt->type) 
	{

	case UART_RX_RDY:														//Some data was received and receive timeout occurred (if RX timeout is enabled) or when the receive buffer is full
		LOG_INF("UART_RX_RDY");
		buf = CONTAINER_OF(evt->data.rx.buf, struct uart_data_t, data);
		buf->len += evt->data.rx.len;

		  //LOG_INF("data len:%d",evt->data.rx.len);
		//   dataReceived = true;

		if (disable_req) {
			return;
		}
		
		
		// if ((evt->data.rx.buf[buf->len - 2] == ']') ||
		//     (evt->data.rx.buf[buf->len - 1] == ']')) {
		if(buf->len < 814){
			
			dataReceived = true;
			buf->data[buf->len] = '\0';
			new_process_uart_data(buf->data);

			// if (server_connect() >= 0) {
				// client_get_new_id();
			// }

			// count_pkt+=1;

			memset(buf->data, 0, sizeof(buf->data));
			buf->len = 0;

			// const char *filename = "/SD:/testfile.txt";
			// write_file(filename, data_wifi);
			// read_file(filename);

			disable_req = true;
			uart_rx_disable(uart);
			// buf->len = 0;
			
		}
		break;

	case UART_RX_DISABLED:													//This event is generated whenever receiver has been stopped, disabled or finished its operation (receive buffer filled) and can be enabled again using uart_rx_enable()
		LOG_INF("UART_RX_DISABLED");
		disable_req = false;

		
		// memset(data_wifi, '0', sizeof(data_wifi));
		// data_wifi[0] = 'H';

		// k_work_submit(&storing_func_work);
		// k_sem_give(&storing_func_sem_2);

		// k_sem_give(&storing_func_sem);

		buf = k_malloc(sizeof(*buf));
		if (buf) {
			buf->len = 0;
		} else {
			dataReceived = false;
			// k_free(buf);
			// LOG_WRN("Not able to allocate UART receive buffer");
			k_work_reschedule(&uart_work, UART_WAIT_FOR_BUF_DELAY);
			return;
		}

		

		uart_rx_enable(uart, buf->data, sizeof(buf->data),
			       UART_WAIT_FOR_RX);

		break;

	case UART_RX_BUF_REQUEST:												//Driver requests next buffer for continuous reception
		LOG_INF("UART_RX_BUF_REQUEST");
		buf = k_malloc(sizeof(*buf));
		// buf_temp = k_malloc(sizeof(*buf_temp));
		if (buf) {
			buf->len = 0;
			if(!uart_rx_buf_rsp(uart, buf->data, sizeof(buf->data)))			//sets the new buffer 
			{
				LOG_INF("New buffer set successfully");
			}
			else
			{
				LOG_ERR("There is error: %d",errno);
			}
		} else {
			// LOG_WRN("Not able to allocate UART receive buffer");
		}

		break;

	case UART_RX_BUF_RELEASED:												//Buffer is no longer used by UART driver
		LOG_INF("UART_RX_BUF_RELEASED\n");
		// LOG_INF("LENGTH OF buffer coming: %d",buf->len);
		buf = CONTAINER_OF(evt->data.rx_buf.buf, struct uart_data_t,
				   data);

		
		// if (buf->len > 0) 
		// {
		// 	k_fifo_put(&fifo_uart_rx_data, buf);
		// } 
		// else 
		// {
			k_free(buf);
			LOG_INF("Buffer is released");
		// }
		
		break;

	case UART_TX_ABORTED:												//Transmitting aborted due to timeout or uart_tx_abort() call
		LOG_INF("UART_TX_ABORTED");
		if (!aborted_buf) {
			aborted_buf = (uint8_t *)evt->data.tx.buf;
		}

		aborted_len += evt->data.tx.len;
		buf = CONTAINER_OF(aborted_buf, struct uart_data_t,
				   data);
		uart_tx(uart, &buf->data[aborted_len],
			buf->len - aborted_len, SYS_FOREVER_MS);

		break;
	
	case UART_RX_STOPPED:
        LOG_ERR("UART RX stopped, reason: %d", evt->data.rx_stop.reason);
		// uart_rx_disable(uart);
        uart_rx_enable(uart, buf->data, sizeof(buf->data), 50);
        break;

	default:
		break;
	}
}

static void uart_work_handler(struct k_work *item)
{
	struct uart_data_t *buf;

	buf = k_malloc(sizeof(*buf));
	if (buf) {
		buf->len = 0;
	} else {
		dataReceived = false;
		// LOG_WRN("Not able to allocate UART receive buffer");
		k_work_reschedule(&uart_work, UART_WAIT_FOR_BUF_DELAY);						//Reschedules the delayable work item to be executed after a specified delay.
		return;
	}
	uart_rx_enable(uart, buf->data, sizeof(buf->data), UART_WAIT_FOR_RX);
}

// static bool uart_test_async_api(const struct device *dev)
// {
// 	const struct uart_driver_api *api =
// 			(const struct uart_driver_api *)dev->api;

// 	return (api->callback_set != NULL);
// }

static int uart_init(void)
{
	int err;
	int pos;
	struct uart_data_t *rx;
	struct uart_data_t *tx;

	struct uart_config config;

	if (!device_is_ready(uart)) {
		return -ENODEV;
	}

	// int ret = uart_config_get(uart, &config);
    // if (ret < 0) {
    //     LOG_ERR("Failed to get UART configuration, ret: %d", ret);
    //     return ret;
    // }

	// config.stop_bits = UART_CFG_STOP_BITS_1;
	// // config.flow_ctrl = UART_CFG_FLOW_CTRL_NONE;

	// ret = uart_configure(uart, &config);
    // if (ret < 0) {
    //     LOG_ERR("Failed to configure UART, ret: %d", ret);
    //     return ret;
    // }

	if (IS_ENABLED(CONFIG_USB_DEVICE_STACK)) {
		err = usb_enable(NULL);
		if (err && (err != -EALREADY)) {
			LOG_ERR("Failed to enable USB");
			return err;
		}
	}

	rx = k_malloc(sizeof(*rx));
	if (rx) {
		rx->len = 0;
	} else {
		return -ENOMEM;
	}

	k_work_init_delayable(&uart_work, uart_work_handler);			//initialises a delayable work item which will be executed after a specific delay, This function prepares the work item to be added to a workqueue for execution


	// if (IS_ENABLED(CONFIG_BT_NUS_UART_ASYNC_ADAPTER) && !uart_test_async_api(uart)) {
	// 	/* Implement API adapter */
	// 	uart_async_adapter_init(async_adapter, uart);
	// 	uart = async_adapter;
	// }
	// if (!uart_test_async_api(uart)) {
    //         /* Implement API adapter */
    //         uart_async_adapter_init(async_adapter, uart);
    //         uart = async_adapter;
    // }

	//registering the callback function, uart_cb is the event handler which will handle the callback functionality
	err = uart_callback_set(uart, uart_cb, NULL);
	if (err) {
		k_free(rx);
		LOG_ERR("Cannot initialize UART callback");
		return err;
	}

	if (IS_ENABLED(CONFIG_UART_LINE_CTRL)) {
		LOG_INF("Wait for DTR");
		while (true) {
			uint32_t dtr = 0;

			uart_line_ctrl_get(uart, UART_LINE_CTRL_DTR, &dtr);
			if (dtr) {
				break;
			}
			/* Give CPU resources to low priority threads. */
			k_sleep(K_MSEC(100));
		}
		LOG_INF("DTR set");
		err = uart_line_ctrl_set(uart, UART_LINE_CTRL_DCD, 1);
		if (err) {
			LOG_WRN("Failed to set DCD, ret code %d", err);
		}
		err = uart_line_ctrl_set(uart, UART_LINE_CTRL_DSR, 1);
		if (err) {
			LOG_WRN("Failed to set DSR, ret code %d", err);
		}
	}
	
	// memset(rx->data, 0, sizeof(rx->data));

	// uart_rx_disable(uart);

	//to start receiving we will calll the below func
	err = uart_rx_enable(uart, rx->data, sizeof(rx->data), 50);
	if (err) {
		LOG_ERR("Cannot enable uart reception (err: %d)", err);
		/* Free the rx buffer only because the tx buffer will be handled in the callback */
		k_free(rx);
	}
	

	return err;
}

/* Setup TLS options on a given socket */
// int tls_setup(int fd)
// {
// 	int err;
// 	int verify;

// 	/* Security tag that we have provisioned the certificate with */
// 	const sec_tag_t tls_sec_tag[] = {
// 		TLS_SEC_TAG,
// 	};

// 	/* Set up TLS peer verification */
// 	enum {
// 		NONE = 0,
// 		OPTIONAL = 1,
// 		REQUIRED = 2,
// 	};

// 	verify = REQUIRED;

// 	err = setsockopt(fd, SOL_TLS, TLS_PEER_VERIFY, &verify, sizeof(verify));
// 	if (err) {
// 		printk("Failed to setup peer verification, err %d\n", errno);
// 		return err;
// 	}

// 	/* Associate the socket with the security tag
// 	 * we have provisioned the certificate with.
// 	 */
// 	err = setsockopt(fd, SOL_TLS, TLS_SEC_TAG_LIST, tls_sec_tag, sizeof(tls_sec_tag));
// 	if (err) {
// 		printk("Failed to setup TLS sec tag, err %d\n", errno);
// 		return err;
// 	}

// 	err = setsockopt(fd, SOL_TLS, TLS_HOSTNAME, HTTPS_HOSTNAME, sizeof(HTTPS_HOSTNAME) - 1);
// 	if (err) {
// 		printk("Failed to setup TLS hostname, err %d\n", errno);
// 		return err;
// 	}
// 	return 0;
// }

// decides whether the wifi connection is a success or failure
static void net_mgmt_event_handler(struct net_mgmt_event_callback *cb, uint32_t mgmt_event,
				   struct net_if *iface)
{
	if ((mgmt_event & EVENT_MASK) != mgmt_event) {
		return;
	}
	if (mgmt_event == NET_EVENT_L4_CONNECTED) {
		LOG_INF("Network connected");
		connected_wifi = true;

		
		// dk_set_led_on(DK_LED1);
		k_sem_give(&run_app);
		return;
	}
	if (mgmt_event == NET_EVENT_L4_DISCONNECTED) {
		// if (connected_wifi == false) {
		// 	LOG_INF("Waiting for network to be connected");
		// } else {
			// dk_set_led_off(DK_LED1);
		LOG_INF("Network disconnected");
		connected_wifi = false;
		// }
		k_sem_reset(&run_app);
		return;
	}

	cmd_wifi_status();
	
	
}

static int wifi_args_to_params(struct wifi_connect_req_params *params)
{

	/* STEP 7.1 Populate the SSID and password */
	params->ssid = CONFIG_WIFI_CREDENTIALS_STATIC_SSID;
	params->ssid_length = strlen(params->ssid);

	params->psk = CONFIG_WIFI_CREDENTIALS_STATIC_PASSWORD;
	params->psk_length = strlen(params->psk);

	/* STEP 7.2 - Populate the rest of the relevant members */
	params->channel = WIFI_CHANNEL_ANY;
	params->security = WIFI_SECURITY_TYPE_PSK;
	params->mfp = WIFI_MFP_OPTIONAL;
	params->timeout = SYS_FOREVER_MS;
	params->band = WIFI_FREQ_BAND_UNKNOWN;

	return 0;
}

static int cmd_wifi_status(void)
{
        struct net_if *iface = net_if_get_default();
        struct wifi_iface_status status = {0};

        if (net_mgmt(NET_REQUEST_WIFI_IFACE_STATUS, iface, &status,
                     sizeof(struct wifi_iface_status)))
        {
                LOG_INF("Status request failed");

                return -ENOEXEC;
        }

        LOG_INF("==================");
        LOG_INF("State: %s", wifi_state_txt(status.state));

        if (status.state >= WIFI_STATE_ASSOCIATED)
        {
                uint8_t mac_string_buf[sizeof("xx:xx:xx:xx:xx:xx")];

                LOG_INF("Interface Mode: %s",
                        wifi_mode_txt(status.iface_mode));
                LOG_INF("Link Mode: %s",
                        wifi_link_mode_txt(status.link_mode));
                LOG_INF("SSID: %-32s", status.ssid);
                LOG_INF("BSSID: %s",
                        net_sprint_ll_addr_buf(
                            status.bssid, WIFI_MAC_ADDR_LEN,
                            mac_string_buf, sizeof(mac_string_buf)));
                LOG_INF("Band: %s", wifi_band_txt(status.band));
                LOG_INF("Channel: %d", status.channel);
                LOG_INF("Security: %s", wifi_security_txt(status.security));
                LOG_INF("MFP: %s", wifi_mfp_txt(status.mfp));
                LOG_INF("RSSI: %d", status.rssi);
        }
        return 0;
}


static int server_resolve(void)
{
	int err;
	struct addrinfo *result;
	struct addrinfo hints = {.ai_family = AF_INET, .ai_socktype = SOCK_STREAM};

	err = getaddrinfo(HTTPS_HOSTNAME, HTTPS_PORT, &hints, &result);
	if (err != 0) {
		LOG_ERR("getaddrinfo failed, err: %d, %s", err, gai_strerror(err));
		return -EIO;
	}

	if (result == NULL) {
		LOG_ERR("Error, address not found");
		return -ENOENT;
	}

	struct sockaddr_in *server4 = ((struct sockaddr_in *)&server);
	server4->sin_addr.s_addr = ((struct sockaddr_in *)result->ai_addr)->sin_addr.s_addr;
	server4->sin_family = AF_INET;
	server4->sin_port = ((struct sockaddr_in *)result->ai_addr)->sin_port;

	char ipv4_addr[NET_IPV4_ADDR_LEN];
	inet_ntop(AF_INET, &server4->sin_addr.s_addr, ipv4_addr, sizeof(ipv4_addr));
	LOG_INF("IPv4 address of HTTP server found %s", ipv4_addr);

	freeaddrinfo(result);


	return 0;
}

static int setup_credentials(void)
{
	
	/* STEP 4.2 - Add the credential to the device */
	int err = tls_credential_add(HTTP_TLS_SEC_TAG, TLS_CREDENTIAL_CA_CERTIFICATE, cert,
				 sizeof(cert));
	if (err < 0) {
		LOG_ERR("Failed to add TLS credentials, err: %d", err);
		return err;
	}

	// err = tls_credential_add(HTTP_TLS_SEC_TAG, TLS_CREDENTIAL_PRIVATE_KEY, private_key, sizeof(private_key));
    // if (err < 0) {
    //     printk("Failed to add private key: %d\n", err);
    //     return;
    // }

	return 0;
}


int server_connect(void)
{
	int err;
	sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TLS_1_2);
	if (sock < 0) {
		LOG_ERR("Failed to set up HTTPS socket, err: %d, %s", errno, strerror(errno));
		return -errno;
	}

	/* STEP 5.1 - Configure the socket with the security tag for the certificate */
	sec_tag_t sec_tag_opt[] = {
		HTTP_TLS_SEC_TAG,
	};
	err = setsockopt(sock, SOL_TLS, TLS_SEC_TAG_LIST, sec_tag_opt, sizeof(sec_tag_opt));
	if (err < 0) {
		LOG_ERR("Failed to set TLS security TAG list, err: %d", errno);
		(void)close(sock);
		return -errno;
	}
	/* STEP 5.2 - Configure the socket with the hostname of the HTTP server */
	err = setsockopt(sock, SOL_TLS, TLS_HOSTNAME, HTTPS_HOSTNAME, sizeof(HTTPS_HOSTNAME));
	if (err < 0) {
		LOG_ERR("Failed to set TLS_HOSTNAME option. Err: %d", errno);
		(void)close(sock);
		return -errno;
	}

	err = connect(sock, (struct sockaddr *)&server, sizeof(struct sockaddr_in));
	if (err < 0) {
		LOG_ERR("Connecting to server failed, err: %d, %s", errno, strerror(errno));
		return -errno;
	}
	
	LOG_INF("Connected to server");
	return 0;
}

static void response_cb(struct http_response *rsp, enum http_final_call final_data, void *user_data)
{
	/* STEP 9 - Define the callback function to print the body */
        LOG_INF("Response status: %s", rsp->http_status);

        if (rsp->body_frag_len > 0) {
                char body_buf[rsp->body_frag_len];
                strncpy(body_buf, rsp->body_frag_start, rsp->body_frag_len);
                body_buf[rsp->body_frag_len]='\0';
                LOG_INF("Received: %s", body_buf);
        } 

        LOG_INF("Closing socket: %d", sock);
        close(sock);

		// server_connect();

}

static void client_id_cb(struct http_response *rsp, enum http_final_call final_data,
			 void *user_data)
{
	/* STEP 6.1 - Log the HTTP response status */
        LOG_INF("Response status: %s", rsp->http_status);

	/* STEP 6.2 - Retrieve and format the client ID */
        char client_id_buf_tmp[CLIENT_ID_SIZE+1];
        strncpy(client_id_buf_tmp, rsp->body_frag_start, CLIENT_ID_SIZE);
        client_id_buf_tmp[CLIENT_ID_SIZE]='\0';
        client_id_buf[0]='/';
        strcat(client_id_buf,client_id_buf_tmp);

        LOG_INF("Successfully acquired client ID: %s", client_id_buf);
	/* STEP 6.3 - Close the socket */
        LOG_INF("Closing socket: %d", sock);
        close(sock);

}

static int read_json_file(const char *filename, char **out_buffer, size_t *out_size) {
    struct fs_file_t file;
    fs_file_t_init(&file);

    int res = fs_open(&file, filename, FS_O_READ);
    if (res) {
        printk("Error opening file %s [%d]\n", filename, res);
        return res;
    }

    res = fs_seek(&file, 0, FS_SEEK_END);
    if (res) {
        printk("Error seeking to end of file %s [%d]\n", filename, res);
        fs_close(&file);
        return res;
    }

    ssize_t file_size = fs_tell(&file);
    if (file_size < 0) {
        printk("Error telling file position %s [%zd]\n", filename, file_size);
        fs_close(&file);
        return file_size;
    }

    res = fs_seek(&file, 0, FS_SEEK_SET);
    if (res) {
        printk("Error seeking to start of file %s [%d]\n", filename, res);
        fs_close(&file);
        return res;
    }

    *out_buffer = k_malloc(file_size + 1);
    if (*out_buffer == NULL) {
        printk("Error allocating buffer\n");
        fs_close(&file);
        return -ENOMEM;
    }

    ssize_t bytes_read = fs_read(&file, *out_buffer, file_size);
    if (bytes_read < 0) {
        printk("Error reading from file %s [%zd]\n", filename, bytes_read);
        k_free(*out_buffer);
        fs_close(&file);
        return bytes_read;
    }

    (*out_buffer)[file_size] = '\0';  // Null-terminate the string
    *out_size = file_size;

    LOG_INF("Read total %zu bytes from file %s\n", file_size, filename);

	k_free(*out_buffer);

    fs_close(&file);
    return 0;

}

static int client_http_put(void)
{
	/* STEP 7 - Define the function to send a PUT request to the HTTP server */
        int err = 0;
        // int bytes_written;
		char *buffer;
		size_t bytes_read;

		// Read the JSON file
		err = read_json_file(full_filename, &buffer, &bytes_read);
		if (err) {
			return err;
    	}


		// // Create content length header
		char content_length_header[50];
		snprintf(content_length_header, sizeof(content_length_header), "Content-Length: %zu\r\n", bytes_read);

        const char *headers[] = {
					"Connection: close\r\n",
					// "Host: ultrahuman-home.storage.googleapis.com\r\n",
					"Content-Type: application/json\r\n",
					content_length_header,
					"Authorization: Bearer ya29.a0AXooCgv9fyEKnKiKrB-h-2TVJXjF8dm_JdkpZcuQfgRag3_FksNuqb9kOtSkkuO2cgPPchtTacCemmdPNy6jEZUTEzOh0RWAu9QUsADtMZwgTensRkYEOR48S_4ryyGWov9vcuAEknQVKWir2wHcuyfvmX65vKKDK3XGaCgYKAYMSARISFQHGX2Mit4SeT3TzbjreQmD3gWdOFQ0171\r\n",
					NULL
				};

        struct http_request req;
        memset(&req, 0, sizeof(req));

		char url[256];
    	snprintf(url, sizeof(url), "/%s", filename);
		// const char *msg = "{"
        //                     "  \"temperature\": \"74\","
        //                     "  \"humidity\": \"35\""
        //                     "}";
		// LOG_INF("JSON data: %s",msg);

        // char buffer[12] = {0};
        // bytes_written = snprintf(buffer, 12, "%d", counter);
        // if (bytes_written < 0){
        //         LOG_INF("Unable to write to buffer, err: %d", bytes_written);
        //         return bytes_written;
        // }
        req.header_fields = headers;
        req.method = HTTP_PUT;
        req.url = url;
        req.host = HTTPS_HOSTNAME;
        req.protocol = "HTTP/1.1";
        req.payload = buffer;
        req.payload_len = bytes_read;
        req.response = response_cb;
        req.recv_buf = recv_buf;
        req.recv_buf_len = sizeof(recv_buf);

        // LOG_INF("HTTP PUT request: %s", buffer);
		LOG_INF("HTTP PUT request with JSON payload of size: %zu", bytes_read);
        err = http_client_req(sock, &req, 5000, NULL);
        if (err < 0) {
                LOG_ERR("Failed to send HTTP PUT request %s, err: %d", buffer, err);
        }
		else
		{
			http_req_success_flag = 1;
		}

		k_free(buffer); // Free the allocated memory

        return err;

}

static int client_http_get(void)
{
	/* STEP 8 - Define the function to send a GET request to the HTTP server */
        int err = 0;
        const char *headers[] = {"Connection: close\r\n", NULL};

        struct http_request req;
        memset(&req, 0, sizeof(req));

        req.header_fields = headers;
        req.method = HTTP_GET;
        req.url = "/data";
        req.host = HTTPS_HOSTNAME;
        req.protocol = "HTTP/1.1";
        req.response = response_cb;
        req.recv_buf = recv_buf;
        req.recv_buf_len = sizeof(recv_buf);
	
        LOG_INF("HTTP GET request");
        err = http_client_req(sock, &req, 5000, NULL);
        if (err < 0) {
                LOG_ERR("Failed to send HTTP GET request, err: %d", err);
        }

        return err;

}
int count =0;
int client_get_new_id()
{
	int err = 0;
	// LOG_INF("count :%d\n",count++);
	char *buffer = NULL;
		size_t bytes_read;

		// Read the JSON file
		err = read_json_file(full_filename, &buffer, &bytes_read);
		if (err) {
			return err;
    	}


		// // Create content length header
		char content_length_header[50];
		snprintf(content_length_header, sizeof(content_length_header), "Content-Length: %zu\r\n", bytes_read);


	/* STEP 5.1 - Define the structure http_request and fill the block of memory */
        struct http_request req;
        memset(&req, 0, sizeof(req));

		char url[256];
    	snprintf(url, sizeof(url), "/o?uploadType=media&name=%s", filename);

		// char arr[300];
			
			// LOG_INF("Data wifi %s",data_wifi);
		// const char *msg = "{\"temperature\":\"75\",\"humidity\":\"42\"}";
		
		// const char *eg = "{\"Hcho-mg/m3\": 0.01,\"Co-PPM\": 1,\"Co2-PPM\": 500,\"Pm1.0-¦Ìg/m3\": 23,\"Pm10-¦Ìg/m3\": 25,\"Pm25-¦Ìg/m3\": 24,\"Voc-mg/m3\": 0,\"Temp-¡æ\": 72,\"Hum-\": 53,\"Noi-DB\": 56,\"UVA-mw/cm2\": 0.22,\"UVB-mw/cm2\": 0.00,\"UVC-mw/cm2\": 0.00,\"RED\": 9,\"GREEN\": 100,\"BLUE\": 4,\"IR\": 5}";
		// // memcpy(arr, data_wifi, sizeof(arr));
		// const char *eg_1 = "{\"Co2-PPM\": 1000}";
		// strcpy(data_wifi, example);

	/* STEP 5.2 - Populate the http_request structure */
        const char *headers[] = {
					"Connection: close\r\n",
					// "Host: ultrahuman-home.storage.googleapis.com\r\n",
					"Content-Type: application/json\r\n",
					content_length_header,
					"Authorization: Bearer ya29.a0AXooCgsJ6_EVZZnnl9FKzvCMOvB4zZcNof7qXzoo_jbQFSIPRwu4Q7-TiTFvFsinNBmd4ob7FbeUNHLoHqTp73nujTaWZM_DMA_itNkvDwwcHQPyQOmZxrsweUSuYe1Q_xm532o0nFtWQ2b3dS3HvAZL_5u155aDivmQaCgYKAbESARISFQHGX2MiypLTFkwdufz2oePD2NoFbw0171\r\n",
					NULL
				};
        req.header_fields = headers;
        req.method = HTTP_POST;
        req.url = url;
        req.host = HTTPS_HOSTNAME;
        req.protocol = "HTTP/1.1";
		req.payload = buffer;
        req.payload_len = bytes_read;
        req.response = response_cb;
        req.recv_buf = recv_buf;
        req.recv_buf_len = sizeof(recv_buf);

	/* STEP 5.3 - Send the request to the HTTP server */
        // LOG_INF("HTTP POST request: %s",req.payload);
		LOG_INF("Size of msg: %d",req.payload_len);
        err = http_client_req(sock, &req, 5000, NULL);
        if (err < 0) {
                LOG_ERR("Failed to send HTTP PUT request %s, err: %d", buffer, err);
        }
		else
		{
			http_req_success_flag = 1;
			LOG_INF("Req flag value: %d",http_req_success_flag);
		}

		k_free(buffer); // Free the allocated memory
		close(sock);
		
	return err;
}

static void send_req_handler()
{
	// if (has_changed & DK_BTN1_MSK && button_state & DK_BTN1_MSK) {
	if (server_connect() >= 0) {
		LOG_INF("Implemented the function in send_req_handler");
		client_http_put();
	// 	// counter++;
	}
        // } else if (has_changed & DK_BTN2_MSK && button_state & DK_BTN2_MSK) {
	if (server_connect() >= 0) {
			client_http_get();
	}

	// if (server_connect() >= 0) {
	// 		client_get_new_id();
	// }
        

}

int parse_value(const char *str) {
    int value = 0;
    // Assuming that we find the key, skip the "value:" substring which is 8 characters long
    str += 8; // Skip over "\"value\":"
    // Skip any spaces or additional characters until the start of the number
    while (*str < '0' || *str > '9') {
        str++;
    }
    // Convert the numerical characters to an integer
    while (*str >= '0' && *str <= '9') {
        value = value * 10 + (*str - '0');
        str++;
    }
    return value;
}

void new_process_uart_data(const char *uart_data) {
	SensorData *sensor;
	

	
// *ppm =25;
	sensor = (SensorData *)k_malloc(sizeof(SensorData));
    if (sensor == NULL) {
        // Handle memory allocation failure
        printf("Memory allocation failed\n");
        return 1;
    }
	int tem = 0, humi = 0, noi = 0, p10 = 0, p1 = 0, p25 = 0, co2 = 0, hch = 0, vo = 0, co = 0, red = 0, blue = 0, green = 0, uva = 0, uvb = 0, uvc = 0, ir = 0;
    // Extract temperature

		char *tempPtr = strstr(uart_data, "Temp");
		// LOG_INF("Temp ptr value: %s",tempPtr);
		if (tempPtr) {
			char *valuePtr = strstr(tempPtr, "\"value\":\t");
			// LOG_INF("Temp value: %s",valuePtr);
			if (valuePtr) {
				tem = parse_value(valuePtr);
			}
		}
    // Extract humidity
		char *humPtr = strstr(uart_data, "Hum");
		if (humPtr) {
			char *valuePtr = strstr(humPtr, "\"value\":\t");
			if (valuePtr) {
				humi = parse_value(valuePtr);
			}
		}
    // Extract noise
		char *noisePtr = strstr(uart_data, "Noi-DB");
		if (noisePtr) {
			char *valuePtr = strstr(noisePtr, "\"value\":\t");
			if (valuePtr) {
				noi = parse_value(valuePtr);
			}
		}

		char *ppm1Ptr = strstr(uart_data, "Pm1.0");
		if (ppm1Ptr) {
			char *valuePtr = strstr(ppm1Ptr, "\"value\":\t");
			if (valuePtr) {
				p1 = parse_value(valuePtr);
			}
		}

		char *ppm10Ptr = strstr(uart_data, "Pm10");
		if (ppm10Ptr) {
			char *valuePtr = strstr(ppm10Ptr, "\"value\":\t");
			if (valuePtr) {
				p10 = parse_value(valuePtr);
			}
		}

		char *ppm25Ptr = strstr(uart_data, "Pm25");
		if (ppm25Ptr) {
			char *valuePtr = strstr(ppm25Ptr, "\"value\":\t");
			if (valuePtr) {
				p25 = parse_value(valuePtr);
			}
		}

		char *co_2ptr = strstr(uart_data, "Co2");
		if (co_2ptr) {
			char *valuePtr = strstr(co_2ptr, "\"value\":\t");
			if (valuePtr) {
				co2 = parse_value(valuePtr);
			}
		}

		char *hcho_ptr = strstr(uart_data, "Hcho");
		if (hcho_ptr) {
			char *valuePtr = strstr(hcho_ptr, "\"value\":\t");
			if (valuePtr) {
				hch = parse_value(valuePtr);
			}
		}

		char *vocptr = strstr(uart_data, "Voc");
		if (vocptr) {
			char *valuePtr = strstr(vocptr, "\"value\":\t");
			if (valuePtr) {
				vo = parse_value(valuePtr);
			}
		}

		char *coptr = strstr(uart_data, "Co");
		if (coptr) {
			char *valuePtr = strstr(coptr, "\"value\":\t");
			if (valuePtr) {
				co = parse_value(valuePtr);
			}
		}

		char *redptr = strstr(uart_data, "RED");
		if (redptr) {
			char *valuePtr = strstr(redptr, "\"value\":\t");
			if (valuePtr) {
				red = parse_value(valuePtr);
			}
		}

		char *greenptr = strstr(uart_data, "Green");
		if (greenptr) {
			char *valuePtr = strstr(greenptr, "\"value\":\t");
			if (valuePtr) {
				green = parse_value(valuePtr);
			}
		}

		char *blueptr = strstr(uart_data, "Bulle");
		if (blueptr) {
			char *valuePtr = strstr(blueptr, "\"value\":\t");
			if (valuePtr) {
				blue = parse_value(valuePtr);
			}
		}

		char *irptr = strstr(uart_data, "Ir");
		if (irptr) {
			char *valuePtr = strstr(irptr, "\"value\":\t");
			if (valuePtr) {
				ir = parse_value(valuePtr);
			}
		}

		char *uvaptr = strstr(uart_data, "UVA");
		if (uvaptr) {
			char *valuePtr = strstr(uvaptr, "\"value\":\t");
			if (valuePtr) {
				uva = parse_value(valuePtr);
			}
		}

		char *uvbptr = strstr(uart_data, "UVB");
		if (uvbptr) {
			char *valuePtr = strstr(uvbptr, "\"value\":\t");
			if (valuePtr) {
				red = parse_value(valuePtr);
			}
		}

		char *uvcptr = strstr(uart_data, "UVC");
		if (uvcptr) {
			char *valuePtr = strstr(uvcptr, "\"value\":\t");
			if (valuePtr) {
				red = parse_value(valuePtr);
			}
		}
		sensor->prodotem = tem;
		sensor->prodohumi = humi;
		sensor->prodonoi = noi;
		sensor->prodop1 = p1;
		sensor->prodop10 = p10;
		sensor->prodop25 = p25;
		sensor->prodoco2 =co2;
		sensor->prodohch = hch;
		sensor->prodovo = vo;
		sensor->prodoco = co;
		sensor->prodored = red;
		sensor->prodoblue = blue;
		sensor->prodogreen = green;
		sensor->prodoUVa = uva;
		sensor->prodoUVb = uvb;
		sensor->prodoUVc = uvc;
		sensor->prodoir = ir;

		// if(tem!=0 && humi!=0){
		// 	snprintf(data_wifi, sizeof(data_wifi),"{\"Hcho-mg/m3\": %.2f,\"Co-PPM\": %d,\"Co2-PPM\": %d,\"Pm1.0-umg/m3\": %d,\"Pm10-umg/m3\": %d,\"Pm25-umg/m3\": %d,\"Voc-mg/m3\": %d,\"Temp-F\": %d,\"Hum-\": %d,\"Noi-DB\": %d,\"UVA-mw/cm2\": %.2f,\"UVB-mw/cm2\": %.2f,\"UVC-mw/cm2\": %.2f,\"RED\": %d,\"GREEN\": %d,\"BLUE\": %d,\"IR\": %d}",
		// 		(float)sensor->prodohch,
		// 		sensor->prodoco,
		// 		sensor->prodoco2,
		// 		sensor->prodop1,
		// 		sensor->prodop10,
		// 		sensor->prodop25,
		// 		sensor->prodovo,
		// 		sensor->prodotem,
		// 		sensor->prodohumi,
		// 		sensor->prodonoi,
		// 		(sensor->prodoUVa/100),
		// 		(sensor->prodoUVb/100),
		// 		(sensor->prodoUVc/100),
		// 		sensor->prodored,
		// 		sensor->prodogreen,
		// 		sensor->prodoblue,
		// 		sensor->prodoir);

			// strcpy(server_data,data_wifi);
			newtime = nrf_cal_get_time_calibrated();
		
			if(tem!=0 && humi!=0){
			snprintf(data_wifi, sizeof(data_wifi),"{\"Timestamp\": %i,\"Hcho-mg/m3\": %.2f,\"Co-PPM\": %d,\"Co2-PPM\": %d,\"Pm1.0-umg/m3\": %d,\"Pm10-umg/m3\": %d,\"Pm25-umg/m3\": %d,\"Voc-mg/m3\": %d,\"Temp-F\": %d,\"Hum-\": %d,\"Noi-DB\": %d,\"UVA-mw/cm2\": %.2f,\"UVB-mw/cm2\": %.2f,\"UVC-mw/cm2\": %.2f,\"RED\": %d,\"GREEN\": %d,\"BLUE\": %d,\"IR\": %d}",
				(int)newtime,
				(float)sensor->prodohch,
				sensor->prodoco,
				sensor->prodoco2,
				sensor->prodop1,
				sensor->prodop10,
				sensor->prodop25,
				sensor->prodovo,
				sensor->prodotem,
				sensor->prodohumi,
				sensor->prodonoi,
				(sensor->prodoUVa/100),
				(sensor->prodoUVb/100),
				(sensor->prodoUVc/100),
				sensor->prodored,
				sensor->prodogreen,
				sensor->prodoblue,
				sensor->prodoir);
			// // LOG_INF("m_last_calibrate_time: %ld",m_last_calibrate_time);
			// LOG_INF("Newtime: %ld",(long)newtime);
			// k_sleep(K_SECONDS(1));
			// local_time_now= localtime(&newtime);
			// LOG_INF("Localtime: %i",(int)local_time_now);
			// Get the current time as epoch time
    		// time_t epoch_time = time(NULL);
			// LOG_INF("Eposch: %ld",epoch_time);
			// char current_time[9];
			// snprintf(current_time, sizeof(current_time), "%i",(int)newtime);
			// // snprintf(current_time, sizeof(current_time), "%02d:%02d:%02d",local_time_now->tm_hour,local_time_now->tm_min,local_time_now->tm_sec);
			// memmove(data_wifi + strlen(current_time) + 1, data_wifi, strlen(data_wifi) + 1);
			// memcpy(data_wifi, current_time, strlen(current_time));
			// data_wifi[strlen(current_time)] = ' ';

			
			LOG_INF("The data from prodo is: %s\r\n",data_wifi);

		}

		// strcpy(server_data,data_wifi);
		
		// client_get_new_id();
		// if (client_get_new_id() < 0) {
		// 	LOG_ERR("Failed to get client ID");
		// 	return 0;
		// }
		k_free(sensor);

		// LOG_INF("Size of data wifi string: %d",strlen(data_wifi));
		// write_file(filename, data_wifi);
		
}	

// void sd_write_thread(struct k_work *work)
// {
//     while (1) {
//         k_sem_take(&sd_write_sem, K_FOREVER);

// 		k_mutex_lock(&file_mutex, K_FOREVER);

//         write_file(filename, data_wifi);
// 		// is_sd_card_empty(filename);
// 		k_mutex_unlock(&file_mutex);
		
// 		k_sleep(K_MSEC(1));

//     }
// }

// void sd_read_thread(struct k_work *work)
// {
//     while (1) {
//         k_sem_take(&sd_read_sem, K_FOREVER);
		
//         k_mutex_lock(&file_mutex, K_FOREVER);
//         read_file(filename);
// 		// clear_file(filename);
// 		// write_file(filename, data_wifi);
		
// 		k_mutex_unlock(&file_mutex);
		
// 		k_sleep(K_MSEC(1));
//     }
// }

// void sd_clear_thread(struct k_work *work)
// {
// 	while (1) {
//         k_sem_take(&sd_clear_sem, K_FOREVER);
		
//         k_mutex_lock(&file_mutex, K_FOREVER);
//         clear_file(filename);
		
// 		k_mutex_unlock(&file_mutex);
		
// 		k_sleep(K_MSEC(1));
//     }
// }
// void storing_func_thread_1(struct k_work *work){
// 	// while(1){
// 		// k_sem_take(&storing_func_sem_1, K_FOREVER);

// 		// k_mutex_lock(&file_mutex, K_FOREVER);
// 		storing_functionality_1();
// 		// k_mutex_unlock(&file_mutex);
		
// 		// k_sleep(K_MSEC(1));

// 	// }
// }


// void storing_functionality_1()
// {
// 	if(connected_wifi == true && dataReceived  == true)				// if wifi is connected and wifi switch is also ON and data is coming from prodo		
// 	{
// 		LOG_INF("data received: %d",dataReceived);
// 		if(!(is_sd_card_empty(filename)))									// if sd card is empty send the received data directly to cloud
// 		{
// 			// k_sem_give(&sd_write_sem);
// 			// k_sem_give(&sd_read_sem);
// 			cloud_flag = send_data_from_prodo;
// 			write_file(filename, data_wifi);
// 			LOG_INF("cloud value: %d",cloud_flag);
// 		}
// 		else																// First read the data from the sd card send it to cloud and store the current incoming data into sd card
// 		{
// 			read_file(filename);
			
// 			cloud_flag = 2;
			
// 			// clear_file(filename);
// 			write_file(filename, data_wifi);
// 			LOG_INF("cloud value: %d",cloud_flag);
// 		}
// 	}
// 	else if(connected_wifi == true && dataReceived == false)			// if wifi is connected and wifi switch is also ON and data is coming from prodo
// 	{
// 		LOG_INF("data received: %d",dataReceived);
// 		if(!(is_sd_card_empty(filename)))									// No data present in the sd card and no incoming data also
// 		{
// 			cloud_flag = 3;
// 			LOG_INF("cloud value: %d",cloud_flag);
// 		}
// 		else																// if sd card is not empty read the data from sd card and send it to cloud
// 		{
// 			read_file(filename);
			
// 			cloud_flag = 4;
			
// 			// clear_file(filename);
// 			LOG_INF("cloud value: %d",cloud_flag);
// 		}
// 	}
// 	else if(!connected_wifi && dataReceived == true)				// if connection goes off or wifi switch is off and data is coming than store it in sd card
// 	{
// 		LOG_INF("data received: %d",dataReceived);
// 		write_file(filename, data_wifi);
// 		cloud_flag = wifi_disconnected;
// 		LOG_INF("cloud value: %d",cloud_flag);
// 	}
	
// }	

static int write_wav_header(const char *filename, uint32_t pcm_data_size) {
    struct wav_header header = {
        .riff_header = "RIFF",
        .wav_size = pcm_data_size + sizeof(struct wav_header) - 8,
        .wave_header = "WAVE",
        .fmt_header = "fmt ",
        .fmt_chunk_size = 16,
        .audio_format = 1,
        .num_channels = 2,
        .sample_rate = 16000,
        .byte_rate = 16000 * 2 * 2,
        .block_align = 2 * 2,
        .bits_per_sample = 16,
        .data_header = "data",
        .data_bytes = pcm_data_size
    };

    struct fs_file_t file;
    fs_file_t_init(&file);

    int res = fs_open(&file, filename, FS_O_CREATE | FS_O_WRITE);
    if (res) {
        printk("Error opening file %s [%d]\n", filename, res);
        return res;
    }

    size_t bytes_written = fs_write(&file, &header, sizeof(header));
    if (bytes_written < sizeof(header)) {
        printk("Error writing WAV header to file %s [%d]\n", filename, bytes_written);
        fs_close(&file);
        return -EIO;
    }

    fs_close(&file);
    return 0;
}

static int update_wav_header(const char *filename, uint32_t pcm_data_size) {
    struct wav_header header = {
        .riff_header = "RIFF",
        .wav_size = pcm_data_size + sizeof(struct wav_header) - 8,
        .wave_header = "WAVE",
        .fmt_header = "fmt ",
        .fmt_chunk_size = 16,
        .audio_format = 1,
        .num_channels = 2,
        .sample_rate = 16000,
        .byte_rate = 16000 * 2 * 2,
        .block_align = 2 * 2,
        .bits_per_sample = 16,
        .data_header = "data",
        .data_bytes = pcm_data_size
    };

    struct fs_file_t file;
    fs_file_t_init(&file);

    int res = fs_open(&file, filename, FS_O_WRITE);
    if (res) {
        printk("Error opening file %s [%d]\n", filename, res);
        return res;
    }

    size_t bytes_written = fs_write(&file, &header, sizeof(header));
    if (bytes_written < sizeof(header)) {
        printk("Error writing WAV header to file %s [%d]\n", filename, bytes_written);
        fs_close(&file);
        return -EIO;
    }

    fs_close(&file);
    return 0;
}

void clear_msgq(struct k_msgq *queue) {
    int16_t dummy_buf[PDM_BUF_SIZE];
    int messages_cleared = 0;
    while (k_msgq_get(queue, &dummy_buf, K_NO_WAIT) == 0) {
        // Just consume the messages without processing
        messages_cleared++;
    }
    LOG_INF("Cleared %d messages from the queue", messages_cleared);
}

void audio_callback(int16_t *pcm_buffer, uint16_t size) {
    // Send PCM data to the queue
    // if (k_msgq_put(&pcm_queue, pcm_buffer, K_NO_WAIT) != 0) {
    //     LOG_WRN("PCM buffer overflow, dropping data");
	// 	k_msgq_purge(&pcm_queue);
    //     if (k_msgq_put(&pcm_queue, pcm_buffer, K_NO_WAIT) != 0)
    //     {
    //         LOG_WRN("AGain tried putting msg in queue, not successful");
    //     }
		// int16_t oldest_buffer[PDM_BUF_SIZE];
        // if (k_msgq_get(&pcm_queue, &oldest_buffer, K_NO_WAIT) == 0) {
        //     LOG_INF("Removed oldest data from queue");
        // } else {
        //     LOG_ERR("Failed to remove oldest data from queue");
        //     // break;
        // }
        // clear_msgq(&pcm_queue);  // Clear the queue when overflow happens
        // if (k_msgq_put(&pcm_queue, pcm_buffer, K_NO_WAIT) != 0) {
        //     LOG_ERR("Failed to put PCM data after clearing the queue");
        // } else {
        //     LOG_INF("Successfully put PCM data after clearing the queue");
        // }
    // }
	

    // }
	if (circular_buffer == NULL) {
        // Allocate memory for the circular buffer in one go
        circular_buffer = k_malloc(CIRCULAR_BUFFER_SIZE * PDM_BUF_SIZE * sizeof(int16_t));
        if (circular_buffer == NULL) {
            // Handle memory allocation failure
            printf("Memory allocation for circular buffer failed\n");
            return;
        }
    }
	// memcpy(circular_buffer[head], pcm_buffer, size * sizeof(int16_t));
    // head = (head + 1) % CIRCULAR_BUFFER_SIZE;
    // if (head == tail) {
    //     buffer_full = true;
    //     LOG_WRN("Circular buffer overflow, data will be overwritten");
    // }
    // buffer_ready = true;
	memcpy(circular_buffer + (head * PDM_BUF_SIZE), pcm_buffer, size * sizeof(int16_t));
    
    head = (head + 1) % CIRCULAR_BUFFER_SIZE;
    if (head == tail) {
        buffer_full = true;
        LOG_WRN("Circular buffer overflow, data will be overwritten");
    }
    buffer_ready = true;
	// memcpy(pcm_double_buffer[active_buffer], pcm_buffer, size * sizeof(int16_t));
    // buffer_ready[active_buffer] = true;
    // active_buffer = !active_buffer;  // Switch buffer
}

ISR_DIRECT_DECLARE(pdm_isr_handler) {
    nrfx_pdm_irq_handler();
    ISR_DIRECT_PM();
    return 1;
}

void nrfx_pdm_event_handler(nrfx_pdm_evt_t const *const p_evt) {

    if (p_evt->buffer_requested) {
        nrfx_pdm_buffer_set(pdm_buf, PDM_BUF_SIZE);
    }
    if (p_evt->buffer_released != 0) {
        audio_callback(pdm_buf, PDM_BUF_SIZE);
    }
}

static void pdm_init(void) {
    nrfx_pdm_config_t pdm_config = NRFX_PDM_DEFAULT_CONFIG(CLK_PIN, DIN_PIN);
    nrfx_pdm_init(&pdm_config, nrfx_pdm_event_handler);
}

// void buffer_check_timer_handler(struct k_timer *timer_id) {
//     if (buffer_full && k_msgq_num_used_get(&pcm_queue) < QUEUE_SIZE / 2) {
//         nrfx_pdm_start();
//         buffer_full = false;
//         LOG_INF("Resuming PDM data collection");
//     } else if (!buffer_full && k_msgq_num_used_get(&pcm_queue) >= QUEUE_SIZE) {
//         nrfx_pdm_stop();
//         buffer_full = true;
//         LOG_WRN("PDM buffer full, stopping PDM");
//     }
// }

// void init_buffer_check_timer(void) {
//     k_timer_init(&buffer_check_timer, buffer_check_timer_handler, NULL);
//     k_timer_start(&buffer_check_timer, K_SECONDS(1), K_SECONDS(1));
// }

void for_data_collection(void){
	if(dataReceived == true){
		// LOG_INF("Mutex locking of uart");
		k_mutex_lock(&file_mutex,K_FOREVER);
		// LOG_INF("Mutex locking of uart started");
		write_file(full_filename, data_wifi);
		// Check if the timer flag is set
        if (five_min_over_flag == 1) {
            // Send and delete the current file
            // if (send_file_to_server(filename) == 0) {
            //     delete_file(filename);
            // }
			if (server_connect() >= 0)
			{
				if (client_get_new_id() < 0)
				{
					LOG_ERR("Failed to get client ID");
				}
			}
			close(sock);	
				
			LOG_INF("Executed the req");
			if(http_req_success_flag == 1){
				clear_file(full_filename);
				LOG_INF("Successfully cleared the file");
			}
			// 
			
			

            // Move to the next file
            // current_file_number++;
            // snprintf(filename, MAX_FILENAME_LENGTH, FILENAME_TEMPLATE, (int)newtime);
			generate_filenames();

            // Reset the timer flag and restart the timer
            five_min_over_flag = 0;
            k_timer_start(&five_minute_timer, K_MINUTES(2), K_NO_WAIT); // Restart the timer for the next 5 minutes
        }
		// LOG_INF("Mutex unlocking of uart");
		k_mutex_unlock(&file_mutex);
		// LOG_INF("Mutex unlocking of uart ended");
	}
	// else
	// {
	// 	clear_file(filename);
	// }
}

void sd_card_init(){
	/* raw disk i/o */
    do {
        static const char *disk_pdrv = "SD";
        uint64_t memory_size_mb;
        uint32_t block_count;
        uint32_t block_size;

        if (disk_access_init(disk_pdrv) != 0) {						    //performs disk initialisation
            LOG_ERR("Storage init ERROR!");
            break;
        }

        if (disk_access_ioctl(disk_pdrv,								//Function to get disk parameters
                DISK_IOCTL_GET_SECTOR_COUNT, &block_count)) {
            LOG_ERR("Unable to get sector count");						//DISK_IOCTL_GET_SECTOR_COUNT: Get the number of sectors in the disk
            break;
        }
        LOG_INF("Block count %u", block_count);

        if (disk_access_ioctl(disk_pdrv,
                DISK_IOCTL_GET_SECTOR_SIZE, &block_size)) {				//DISK_IOCTL_GET_SECTOR_SIZE: Get the size of a disk SECTOR in bytes.
            LOG_ERR("Unable to get sector size");
            break;
        }
        printk("Sector size %u\n", block_size);

        memory_size_mb = (uint64_t)block_count * block_size;
        printk("Memory Size(MB) %u\n", (uint32_t)(memory_size_mb >> 20));
    } while (0);

    mp.mnt_point = disk_mount_pt;

    int res = fs_mount(&mp);					//mounts the filesystem on disk

    if (res == FR_OK) {
        printk("Disk mounted.\n");
        lsdir(disk_mount_pt);

    } else {
        printk("Error mounting disk.\n");
    }

}

ssize_t is_sd_card_empty(const char *filename)
{
	struct fs_file_t file;
    fs_file_t_init(&file);

    int res = fs_open(&file, filename, FS_O_READ);
    if (res) {
        LOG_ERR("Error opening file %s [%d]\n", filename, res);
        return true;
    }

    fs_seek(&file, 0, FS_SEEK_END);
    ssize_t file_size = fs_tell(&file);
    fs_close(&file);

	// LOG_INF("File size is: %zd bytes", file_size);

    return file_size;
}

static int lsdir(const char *path)
{
    int res;
    struct fs_dir_t dirp;
    static struct fs_dirent entry;

    fs_dir_t_init(&dirp);								//Initializes the fs_dir_t object; the function needs to be invoked on object before first use with fs_opendir.

    /* Verify fs_opendir() */
    res = fs_opendir(&dirp, path);						//Opens an existing directory specified by the path.
    if (res) {
        printk("Error opening dir %s [%d]\n", path, res);
        return res;
    }

    printk("\nListing dir %s ...\n", path);
    for (;;) {
        /* Verify fs_readdir() */
        res = fs_readdir(&dirp, &entry);					//Reads directory entries of an open directory

        /* entry.name[0] == 0 means end-of-dir */
        if (res || entry.name[0] == 0) {
            break;
        }

        if (entry.type == FS_DIR_ENTRY_DIR) {
            printk("[DIR ] %s\n", entry.name);
        } else {
            printk("[FILE] %s (size = %zu)\n",
                entry.name, entry.size);
        }
    }

    /* Verify fs_closedir() */
    fs_closedir(&dirp);										//Closes an open directory.

    return res;
}


static char *last_written_data = NULL;
static size_t last_written_size = 0;
static int write_file(const char *filename, const char *data)
{
	static char last_written_data[280] = "";
    struct fs_file_t file;
	size_t bytes_written;
    fs_file_t_init(&file);									//Initializes the fs_file_t object; the function needs to be invoked on object before first use with fs_open.

	// k_mutex_lock(&txt_file_mutex, K_FOREVER);

	

	// if (is_all_zero(data)) {
    //     LOG_INF("All values are zero. Skipping write.");
    //     return 0;
    // }


    int res = fs_open(&file, filename, FS_O_CREATE | FS_O_WRITE | FS_O_APPEND);			//Opens or possibly creates a file and associates a stream with it.
    if (res) {
        printk("Error opening file %s [%d]\n", filename, res);
		// k_mutex_unlock(&txt_file_mutex);
        return res;
    }
    
	// int status = strcmp(temp_buf, data_wifi);
	
    // uint32_t start_time = k_cycle_get_32();
    // fs_truncate(&file,0);
    size_t data_len = strlen(data);
	// LOG_INF("Data length: %d",data_len);
	
	// Compare current data with the last written data
    if (strcmp(last_written_data, data) != 0) {
        // Data is different, so write it to the file
        bytes_written = fs_write(&file, data, data_len);  // Attempts to write data_len bytes to the specified file

        if (bytes_written < data_len) {
            printk("Error writing to file %s [%d]\n", filename, bytes_written);
            fs_close(&file);
			// k_mutex_unlock(&txt_file_mutex);
            return -EIO;
        } else {
            LOG_INF("Wrote %zu bytes to file %s\n", data_len, filename);
            // Update the last_written_data buffer
            strncpy(last_written_data, data, sizeof(last_written_data) - 1);
            last_written_data[sizeof(last_written_data) - 1] = '\0';  // Ensure null termination
        }
    } else {
        LOG_INF("Data is the same as the last written data. Skipping write.");
    }

    // uint32_t duration = end_time - start_time;
    // float write_speed = ((float)data_len / duration) * sys_clock_hw_cycles_per_sec();

    // printk("Wrote %zu bytes to file %s\n", data_len, filename);
    // printk("Write duration: %u cycles\n", duration);
    // printk("Write speed: %.2f bytes/second\n", write_speed);

    fs_close(&file);
	// k_mutex_unlock(&txt_file_mutex);													//Flushes the associated stream and closes the file.
    return 0;
}

void connect_wifi()
{
	int err;
	static struct wifi_connect_req_params cnx_params;
	struct net_if *iface = net_if_get_first_wifi();
	if(iface == NULL) {
		LOG_ERR("Returned network interface is NULL");
		return -1;
	}

	wifi_args_to_params(&cnx_params);

	

	err = net_mgmt(NET_REQUEST_WIFI_CONNECT, iface, &cnx_params, sizeof(struct wifi_connect_req_params));
	if (err) {
		LOG_ERR("Connecting to Wi-Fi failed, err: %d", err);
		return ENOEXEC;
	}
}

void disconnect_wifi()
{
	// int ret;

    // // Request to disconnect from Wi-Fi
    // ret = net_mgmt(NET_REQUEST_WIFI_DISCONNECT, iface, NULL, 0);
    // if (ret) {
    //     LOG_ERR("Wi-Fi disconnection failed: %d", ret);
    // } else {
    //     LOG_INF("Wi-Fi disconnected successfully");
    // }
}

void wifi_switch_status(const struct device *dev, struct gpio_callback *cb,
		    uint32_t pins)
{
	if(!gpio_pin_get_dt(&sw2))
	{
		//wifi switch is on
		connect_wifi();
	}
	else
	{
		white_color_off();
		gpio_pin_set_dt(&led1, 1);			//blue led on
		disconnect_wifi();
	}
	// LOG_INF("Wifi switched on");

}

void mic_switch_status(const struct device *dev, struct gpio_callback *cb,
		    uint32_t pins)
{
	if(!gpio_pin_get_dt(&sw1))
	{
		// mic switch is on
		LOG_INF("Mic switch is on");
		nrfx_err_t ret = nrfx_pdm_start();
		if (ret == NRFX_SUCCESS) {
			LOG_INF("Pdm start was successful");
		} else {
			LOG_ERR("Pdm start was NOT successful, error code: %d", ret);
		}
		gpio_pin_set_dt(&led3, 1);				//red led off
	}
	else
	{
		LOG_INF("Mic switch is off");
		nrfx_pdm_stop();
		white_color_off();
		gpio_pin_set_dt(&led3, 0);				//red led on
		
	}
}

void white_color_on(){
	//to glow white light
	gpio_pin_set_dt(&led1, 0);
	gpio_pin_set_dt(&led2, 0);
	gpio_pin_set_dt(&led3, 0);    // startup led on
}

void white_color_off(){
	//to glow white light
	gpio_pin_set_dt(&led1, 1);
	gpio_pin_set_dt(&led2, 1);
	gpio_pin_set_dt(&led3, 1);    // startup led on
}

/* Initialisation of gpio */
int gpio_init() {
    const struct gpio_dt_spec* gpios[] = {&led1, &led2, &led3, &sw1, &sw2};
    int err;

    for (size_t i = 0; i < ARRAY_SIZE(gpios); i++) {
        err = !gpio_is_ready_dt(gpios[i]);
        if (err) {
            return err;
        }
    }

    return 0;
}

/* Confiiguration of gpios as output or input */
int gpio_configure() {
    const struct {											// for all the gpios
        const struct gpio_dt_spec *gpio;
        int flags;
    } gpios[] = {
        {&led1, GPIO_OUTPUT},
        {&led2, GPIO_OUTPUT},
        {&led3, GPIO_OUTPUT},
        {&sw1, GPIO_INPUT | GPIO_PULL_UP},
        {&sw2, GPIO_INPUT | GPIO_PULL_UP}
    };

    const struct {											// only for switches
        const struct gpio_dt_spec *gpio;
        int flags;
    } int_gpios[] = {
        {&sw1, GPIO_INT_EDGE_BOTH},
        {&sw2, GPIO_INT_EDGE_BOTH}
    };

    int err;

    for (size_t i = 0; i < ARRAY_SIZE(gpios); i++) {
        err = gpio_pin_configure_dt(gpios[i].gpio, gpios[i].flags);
        if (err < 0) {
            return err;
        }
    }

    for (size_t i = 0; i < ARRAY_SIZE(int_gpios); i++) {
        err = gpio_pin_interrupt_configure_dt(int_gpios[i].gpio, int_gpios[i].flags);
        if (err < 0) {
            return err;
        }
    }

    return 0;
}



static int write_wav_data(const char *filename, void *data, size_t total_size) {
    struct fs_file_t file;
    fs_file_t_init(&file);

    int res = fs_open(&file, filename, FS_O_CREATE | FS_O_WRITE | FS_O_APPEND);
    if (res) {
        LOG_ERR("Error opening file %s [%d]", filename, res);
        return res;
    }

    size_t bytes_written = 0;
    while (bytes_written < total_size) {
        ssize_t written = fs_write(&file, (uint8_t *)data + bytes_written, total_size - bytes_written);
        if (written < 0) {
            LOG_ERR("Error writing to file %s [%d]", filename, written);
            fs_close(&file);
            return written;
        }
        bytes_written += written;
		// k_msleep(10);
    }

    res = fs_sync(&file);
    if (res != 0) {
        LOG_ERR("Error syncing file %s [%d]", filename, res);
    } else {
        LOG_INF("Successfully wrote %zu bytes to file %s", total_size, filename);
    }

    fs_close(&file);
    return 0;
}

void generate_filenames() {

    // Buffer for the full path
    snprintf(full_filename, MAX_FILENAME_LENGTH, FILENAME_TEMPLATE, (int)newtime);

    // Extract the filename from the full path
    const char *basename = strrchr(full_filename, '/');
    if (basename) {
        basename++;  // Move past the '/'
    } else {
        basename = full_filename;  // If no '/' is found, the filename is the full path
    }

    // Store the basename in the global variable
    strncpy(filename, basename, MAX_FILENAME_LENGTH);
    filename[MAX_FILENAME_LENGTH - 1] = '\0';  // Ensure null termination
}

void pcm_processing_thread(void) {
    // int16_t buffer[PDM_BUF_SIZE];
    uint32_t pcm_data_size = 0;

    while (1) {
        // Receive PCM data from the queue
		// k_mutex_lock(&file_mutex, K_FOREVER);
        // k_msgq_get(&pcm_queue, buffer, K_FOREVER);
		// // LOG_INF("Mutex locking of pdm");
		
		// // LOG_INF("Mutex locking of pdm started");
        // // Write the PCM data to the file
        // int write_res = write_wav_data(wav_filename, buffer, sizeof(buffer));
        // if (write_res != 0) {
        //     LOG_ERR("Failed to write PCM data to file, error code: %d", write_res);
        // } else {
        //     pcm_data_size += sizeof(buffer);
        //     // Update the WAV header periodically
        //     update_wav_header(wav_filename, pcm_data_size);
		// 	// k_msgq_purge(&pcm_queue);
			
        // }
		// // LOG_INF("Mutex unlocking of pdm");
		// k_mutex_unlock(&file_mutex);
		k_mutex_lock(&file_mutex, K_FOREVER);
		// if (buffer_ready) {
            
        //     while (head != tail) {
        //         int16_t *buffer = circular_buffer[tail];
		// 		// memcpy(buffer, circular_buffer[tail], PDM_BUF_SIZE * sizeof(int16_t));
        //         int write_res = write_wav_data(wav_filename, buffer, sizeof(circular_buffer[tail]));
        //         if (write_res != 0) {
        //             LOG_ERR("Failed to write PCM data to file, error code: %d", write_res);
        //         } else {
        //             pcm_data_size += sizeof(circular_buffer[tail]);
        //             update_wav_header(wav_filename, pcm_data_size);
        //         }
		// 		// Clear the buffer after processing
        //         // memset(circular_buffer[tail], 0, PDM_BUF_SIZE * sizeof(int16_t));
		// 		// memset(buffer, 0, PDM_BUF_SIZE * sizeof(int16_t));
        //         tail = (tail + 1) % CIRCULAR_BUFFER_SIZE;
        //     }
        //     buffer_ready = false;
            
        // }
		if (buffer_ready) {
            while (head != tail) {
                int16_t *buffer = circular_buffer + (tail * PDM_BUF_SIZE);
                int write_res = write_wav_data(wav_filename, buffer, PDM_BUF_SIZE * sizeof(int16_t));
                if (write_res != 0) {
                    LOG_ERR("Failed to write PCM data to file, error code: %d", write_res);
                } else {
                    pcm_data_size += PDM_BUF_SIZE * sizeof(int16_t);
                    update_wav_header(wav_filename, pcm_data_size);
                }
                tail = (tail + 1) % CIRCULAR_BUFFER_SIZE;
				// buffer_full = false;
            }
            buffer_ready = false;
        }
		k_mutex_unlock(&file_mutex);
        k_sleep(K_MSEC(10)); // Adjust the sleep duration as needed

		
		// LOG_INF("Mutex unlocking of pdm ended");
		// if (buffer_ready[!active_buffer]) {
        //     // Write the PCM data from the inactive buffer to the file
        //     k_mutex_lock(&file_mutex, K_FOREVER);
        //     int write_res = write_wav_data(wav_filename, pcm_double_buffer[!active_buffer], sizeof(pcm_double_buffer[!active_buffer]));
        //     if (write_res != 0) {
        //         LOG_ERR("Failed to write PCM data to file, error code: %d", write_res);
        //     } else {
        //         pcm_data_size += sizeof(pcm_double_buffer[!active_buffer]);
        //         // Update the WAV header periodically
        //         update_wav_header(wav_filename, pcm_data_size);
		// 		memset(pcm_double_buffer[!active_buffer], 0, DOUBLE_BUFFER_SIZE * sizeof(int16_t));
        //     }
            

        //     buffer_ready[!active_buffer] = false;  // Mark buffer as processed
		// 	k_mutex_unlock(&file_mutex);
        // }

        // k_sleep(K_MSEC(10));  // Small sleep to reduce CPU usage
    }
	// k_free(circular_buffer);
	// k_sleep(K_MSEC(1));
}

// void sd_write_thread(){
// 	char uart_data[sizeof(data_wifi)];

//     while (1) {
//         k_msgq_get(&uart_msgq, &uart_data, K_FOREVER);

//         k_mutex_lock(&txt_file_mutex, K_FOREVER);
//         int res = write_file(filename, uart_data);
//         if (res != 0) {
//             LOG_ERR("Failed to write UART data to file, error code: %d", res);
//         }
//         k_mutex_unlock(&txt_file_mutex);

//         // k_sleep(K_MSEC(10));
//     }
	

// }

// K_THREAD_DEFINE(pcm_thread_id, 4096, pcm_processing_thread, NULL, NULL, NULL, 5, 0, 0);
// K_THREAD_DEFINE(sd_write_tid, 4096, sd_write_thread, NULL, NULL, NULL, 4, 0, 0);




int main(void)
{
	
	// time_t epoch = 1622505600;
    // nrf_cal_set_time(epoch);
	// k_work_init_delayable(&check_cond_work, check_cond);

	
	int err = 0;
	// struct wifi_connect_req_params cnx_params;
	// struct net_if *iface = net_if_get_first_wifi();
	// if(iface == NULL) {
	// 	LOG_ERR("Returned network interface is NULL");
	// 	return -1;
	// }
	

	err = gpio_init();
	if (err < 0) {
		// error();
		LOG_ERR("GPIO not initialised");
	}

	err = gpio_configure();
	if (err < 0) {
		// error();
		LOG_ERR("GPIO not configured");
	}

	white_color_on();
	
	
	// uart_polling_receive();

	
	
	// cJSON_Init();
	

	
	// init_work_items();
	/* Sleep to allow initialization of Wi-Fi driver */
	k_sleep(K_SECONDS(1));

	/* Provision certificates before connecting to the network */
	

	net_mgmt_init_event_callback(&mgmt_cb, net_mgmt_event_handler, EVENT_MASK);
	net_mgmt_add_event_callback(&mgmt_cb);


	// if wifi switch is ON when device is powered on
	if (!gpio_pin_get_dt(&sw2)){
		// initiates the wifi connection
		connect_wifi();
		// LOG_INF("Wifi switch is on at startup");
		// err = net_mgmt(NET_REQUEST_WIFI_CONNECT, iface, &cnx_params, sizeof(struct wifi_connect_req_params));
		// if (err) {
		// 	LOG_ERR("Connecting to Wi-Fi failed, err: %d", err);
		// 	return ENOEXEC;
		// }
		// gpio_pin_set_dt(&led1, 1);				//blue led off
		// white_color_off();
	}
	// LOG_INF("Waiting to connect to Wi-Fi");
	// // initiates the wifi connection
	// err = net_mgmt(NET_REQUEST_WIFI_CONNECT, iface, &cnx_params, sizeof(struct wifi_connect_req_params));
	// if (err) {
	// 	LOG_ERR("Connecting to Wi-Fi failed, err: %d", err);
	// 	return ENOEXEC;
	// }
	// // else {
	// // 	LOG_INF("Connected to wifi");
	// // }
	

	k_sem_take(&run_app, K_FOREVER);

	LOG_INF("################");

	
	// gpio_pin_set_dt(&led2, 1);
	sd_card_init();
	// pdm_init();
	// // Initialize the WAV file with an empty header
    // write_wav_header(wav_filename, 0);

    // IRQ_DIRECT_CONNECT(PDM0_IRQn, 5, pdm_isr_handler, 0);  // Set high priority for PDM interrupt

    // nrfx_err_t ret = nrfx_pdm_start();
    // if (ret == NRFX_SUCCESS) {
    //     LOG_INF("Pdm start was successful");
    // } else {
    //     LOG_ERR("Pdm start was NOT successful, error code: %d", ret);
    // }

	// Check the initial state of the switch
    // if (!gpio_pin_get_dt(&sw1)) {
    //     LOG_INF("Mic switch is on at startup");
    //     nrfx_err_t ret = nrfx_pdm_start();
    //     if (ret == NRFX_SUCCESS) {
    //         LOG_INF("PDM start was successful at startup");
    //         // gpio_pin_set(led_dev, LED2_PIN, 0);  // Turn off red LED
    //     } else {
    //         LOG_ERR("PDM start was NOT successful at startup, error code: %d", ret);
    //     }
	// 	gpio_pin_set_dt(&led3, 1);			// red led off
    // }

	// gpio_init_callback(&sw1_cb_data, mic_switch_status, BIT(sw1.pin));
	// // gpio_init_callback(&sw2_cb_data, wifi_switch_status, BIT(sw2.pin));

	// gpio_add_callback(sw1.port, &sw1_cb_data);


	err = uart_init();
	if (err < 0) {
		// error();
		LOG_ERR("UART not initialised");
	}
	else
	{
		LOG_INF("UART initialised");
	}

	// send_req_handler();
	
	if (server_resolve() != 0) {
		LOG_ERR("Failed to resolve server name");
		return 0;
	}

	if (setup_credentials() != 0) {
		LOG_ERR("Setup credentials failed");
	}

	if (server_connect() != 0) {
		LOG_ERR("Failed to initialize client");
		return 0;
	}
	


	// if (client_get_new_id() < 0) {
	// 	LOG_ERR("Failed to get client ID");
	// 	return 0;
	// }

	// gpio_add_callback(sw2.port, &sw2_cb_data);

	// send_req_handler();
	
	// if (server_resolve() != 0) {
	// 	LOG_ERR("Failed to resolve server name");
	// 	return 0;
	// }

	// if (server_connect() != 0) {
	// 	LOG_ERR("Failed to initialize client");
	// 	return 0;
	// }

	// if (client_get_new_id() < 0) {
	// 	LOG_ERR("Failed to get client ID");
	// 	return 0;
	// }

	

	// Simulate UART data reception and put it in the message queue
    // while (1) {
    //     if (dataReceived) {
    //         k_msgq_put(&uart_msgq, &data_wifi, K_NO_WAIT);
    //         // dataReceived = false;
    //     }
    //     k_sleep(K_SECONDS(1));
    // }
	

	// while(1)
	// {
	// 	if(!gpio_pin_get_dt(&sw1))
	// 	{
	// 		LOG_INF("MIC is on");
	// 	}
	// 	else if(!gpio_pin_get_dt(&sw2))
	// 	{
	// 		LOG_INF("WIFI is on");
	// 	}
	// }
	/* Start the timer with a delay of 300000 ms (5 minutes) and a period of 0 ms */
	// snprintf(filename, MAX_FILENAME_LENGTH, FILENAME_TEMPLATE, (int)newtime);
	generate_filenames();
    k_timer_start(&five_minute_timer, K_MINUTES(2), K_NO_WAIT);

    while (1) {
		// k_sem_give(&sd_write_sem);
		for_data_collection();
		// if (server_connect() >= 0) {
        //     if (client_get_new_id() < 0){
		// 		LOG_ERR("Failed to get client ID");
		// 	}
        // }
		// close(sock);
		k_sleep(K_SECONDS(1));
		
	// // // // 	// log_memory_usage();
	// 	
	// // // // 	// storing_functionality_1();
	// 	k_sleep(K_SECONDS(1));
	// // // // 	// log_memory_usage();
    }

}




// static int read_file(const char *filename)
// {
//     struct fs_file_t file;
//     fs_file_t_init(&file);

//     int res = fs_open(&file, filename, FS_O_READ);
//     if (res) {
//         printk("Error opening file %s [%d]\n", filename, res);
//         return res;
//     }

//     // uint32_t start_time = k_cycle_get_32();
//     char read_buffer[SD_BUFFER_SIZE];
//     ssize_t bytes_read, total_bytes_read = 0;

//     while ((bytes_read = fs_read(&file, read_buffer, sizeof(read_buffer) - 1)) > 0) {			//Reads up to size bytes of data to ptr pointed buffer, returns number of bytes read
//         total_bytes_read += bytes_read;
//         // read_buffer[bytes_read] = '\0'; // Null-terminate the string
//         // printk("Read %zd bytes from file %s: %s\n", bytes_read, filename, read_buffer);
//     }

//     // uint32_t end_time = k_cycle_get_32();
//     if (bytes_read < 0) {
//         printk("Error reading from file %s [%zd]\n", filename, bytes_read);
//         fs_close(&file);
//         return bytes_read;
//     }

//     // res = fs_seek(&file, 0, FS_SEEK_END);
//     // if (res) {
//     //     LOG_ERR("Error seeking to end of file %s [%d]", filename, res);
//     //     fs_close(&file);
//     //     return res;
//     // }

//     // ssize_t file_size = fs_tell(&file);
//     // if (file_size < 0) {
//     //     LOG_ERR("Error telling file position %s [%zd]", filename, file_size);
//     //     fs_close(&file);
//     //     return file_size;
//     // }
    
//     LOG_INF("Read total %zu bytes from file %s\n", total_bytes_read, filename);

// 	return total_bytes_read;

//     // LOG_INF("Size of file %s: %zd bytes", filename, file_size);
//     // uint32_t duration = end_time - start_time;
//     // float read_speed = ((float)total_bytes_read / duration) * sys_clock_hw_cycles_per_sec();

//     // printk("Read total %zu bytes from file %s\n", total_bytes_read, filename);
//     // printk("Read duration: %u cycles\n", duration);
//     // printk("Read speed: %.2f bytes/second\n", read_speed);

//     // fs_unlink(filename);
// 	// lsdir(disk_mount_pt);
//     fs_close(&file);
	
//     return 0;
// }

static int clear_file(const char *filename)
{	
	struct fs_file_t file;
    fs_file_t_init(&file);

    int res = fs_open(&file, filename, FS_O_CREATE | FS_O_WRITE);
    if (res) {
        LOG_ERR("Error opening file %s [%d]\n", filename, res);
        return true;
    }

	// Write zero-length data to the file
    
	// fs_truncate(&file,0);
	fs_unlink(filename);
	fs_close(&file);
	return 0;

}


	

	// time_t epoch = 0;
    // nrf_cal_set_time(epoch);

	// /* STEP 11 - Retrieve the client ID upon connection */



// char* process_uart_data(const char *bufferData) {
//     const char *msg = strstr(bufferData, "[{");

//     // Parse the JSON string
//     cJSON *root = cJSON_Parse(msg);
//     if (root == NULL) {
//         const char *error_ptr = cJSON_GetErrorPtr();
//         if (error_ptr != NULL) {
//             fprintf(stderr, "Error before: %s\n", error_ptr);
//         }
//         return NULL;
//     }
//     // Create a cJSON object to store combined data
//     cJSON *combined_data = cJSON_CreateObject();

//     // Iterate over each item in the array
//     cJSON *item = root->child;
//     while (item != NULL) {
//         // Get "comp" and "value" from the object
//         cJSON *comp = cJSON_GetObjectItemCaseSensitive(item, "comp");
//         cJSON *value = cJSON_GetObjectItemCaseSensitive(item, "value");

//         // Add the value to the combined data using the "comp" value as the key
//         cJSON_AddItemToObject(combined_data, comp->valuestring, cJSON_Duplicate(value, 1));

// 		// Store the next item before deleting the current one
// 		cJSON *nextItem = item->next;

// 		// Delete the current item from the object
// 		cJSON_DeleteItemFromObject(root, item->string);

// 		// Move to the next item
// 		item = nextItem;
//         // // Move to the next item
//         // item = item->next;
//     }
// 	// delete(item);

//     // Print the combined data
//     char *combined_json_str = cJSON_Print(combined_data);
//     // printf("The combined JSON string is: %s\n", combined_json_str);

//     // Remove whitespace characters from the JSON string
//     size_t len = strlen(combined_json_str);
//     char *cleaned_json_str = malloc(len + 1);
//     if (cleaned_json_str == NULL) {
//         cJSON_Delete(root);
//         cJSON_Delete(combined_data);
//         return NULL;  // Memory allocation failed
//     }

//     size_t j = 0;
//     for (size_t i = 0; i < len; i++) {
//         char c = combined_json_str[i];
//         if (c != ' ' && c != '\t' && c != '\n' && c != '\r') {
//             cleaned_json_str[j++] = c;
//         }
//     }
//     cleaned_json_str[j] = '\0';  // Null-terminate the cleaned JSON string

// 	cJSON_free(combined_json_str);
//     // Cleanup cJSON objects
//     cJSON_Delete(root);
//     cJSON_Delete(combined_data);

// 	// LOG_INF("cleaned json: %s",cleaned_json_str);
//     return cleaned_json_str;
// }


